from django.utils.cache import patch_vary_headers
from django.utils import translation


class MinimalLocaleMiddleware(object):
    """
    This is a minimal version of the LocaleMiddleware from Django.
    It only supports setting the current language from sessions.
    This allows the main site to be in one language while the site
    administrators can switch the language, so they don't experience
    problems while editing original database content when this is not
    in the main site's language.
    """

    def process_request(self, request):
        language = get_language_from_request(request)
        translation.activate(language)
        request.LANGUAGE_CODE = translation.get_language()

    def process_response(self, request, response):
        patch_vary_headers(response, ('Accept-Language',))
        if 'Content-Language' not in response:
            response['Content-Language'] = translation.get_language()
        translation.deactivate()
        return response


def get_language_from_request(request):
    from django.conf import settings
    supported = dict(settings.LANGUAGES)

    if hasattr(request, 'session'):
        lang_code = request.session.get('django_language', None)
        if lang_code in supported and lang_code is not None:
            return lang_code
    return settings.LANGUAGE_CODE



VERSION = (0, 1, 1)


def get_version():
    if len(VERSION) > 3 and VERSION[3] != 'final':
        return '%s.%s.%s %s' % (VERSION[0], VERSION[1], VERSION[2], VERSION[3])
    else:
        return '%s.%s.%s' % (VERSION[0], VERSION[1], VERSION[2])


__version__ = get_version()



import sys

from django.core.cache import cache
from django.test import TestCase
from django.utils import translation
from django.contrib.contenttypes.models import ContentType

from datatrans.models import KeyValue, make_digest


if len(sys.argv) > 1 and sys.argv[1] == 'test':
    # We need a model to translate. This is a bit hacky, see
    # http://code.djangoproject.com/ticket/7835
    from django.db import models

    class ModelToTranslate(models.Model):
        message = models.TextField()


class DatatransTests(TestCase):
    def setUp(self):
        self.nl = 'nl'
        self.en = 'en'
        self.message_en = 'Message in English'
        self.message_nl = 'Bericht in het Nederlands'
        self.field = 'message'
        self.instance = ModelToTranslate.objects.create(message=self.message_en)

    def test_default_values(self):
        value = KeyValue.objects.lookup(self.message_en, self.nl, self.instance, self.field)
        self.assertEqual(value, self.message_en)

        kv = KeyValue.objects.get_keyvalue(self.message_en, self.nl, self.instance, self.field)
        kv.value = self.message_nl
        kv.save()

        value = KeyValue.objects.lookup(self.message_en, self.nl, self.instance, self.field)
        self.assertEqual(value, self.message_en)

        kv.edited = True
        kv.save()

        value = KeyValue.objects.lookup(self.message_en, self.nl, self.instance, self.field)
        self.assertEqual(value, self.message_nl)

    def test_cache(self):
        digest = make_digest(self.message_en)
        type_id = ContentType.objects.get_for_model(self.instance.__class__).id
        cache_key = 'datatrans_%s_%s_%s_%s_%s' % (self.nl,
                                                  digest,
                                                  type_id,
                                                  self.instance.id,
                                                  self.field)

        self.assertEqual(cache.get(cache_key), None)

        translation.activate(self.nl)

        kv = KeyValue.objects.get_keyvalue(self.message_en, self.nl, self.instance, self.field)
        self.assertEqual(cache.get(cache_key).value, self.message_en)
        kv.value = self.message_nl
        kv.save()
        kv = KeyValue.objects.get_keyvalue(self.message_en, self.nl, self.instance, self.field)
        self.assertEqual(cache.get(cache_key).value, self.message_nl)
        kv.value = '%s2' % self.message_nl
        kv.save()
        self.assertEqual(cache.get(cache_key).value, '%s2' % self.message_nl)
        kv.delete()
        self.assertEqual(cache.get(cache_key), None)

    def test_fuzzy(self):
        kv = KeyValue.objects.get_keyvalue(self.message_en, self.nl, self.instance, self.field)
        kv.value = self.message_nl
        kv.edited = True
        kv.fuzzy = True
        kv.save()

        value = KeyValue.objects.lookup(self.message_en, self.nl, self.instance, self.field)
        self.assertEqual(value, self.message_nl)



from django.contrib import admin

from datatrans.models import KeyValue


class KeyValueAdmin(admin.ModelAdmin):
    list_display = ('content_type', 'object_id', 'field',
                    'value', 'language', 'edited', 'fuzzy')
    ordering = ('digest', 'language')
    search_fields = ('content_type__app_label', 'content_type__model', 'value',)
    list_filter = ('content_type', 'language', 'edited', 'fuzzy')

admin.site.register(KeyValue, KeyValueAdmin)



import datetime
from django.conf import settings
from django.core.cache import cache
from django.db import models
from django.db.models import signals
from django.db.models.query import QuerySet
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes import generic

from hashlib import sha1


def make_digest(key):
    """Get the SHA1 hexdigest of the given key"""
    return sha1(key.encode('utf-8')).hexdigest()


def _get_cache_keys(self):
    """Get all the cache keys for the given object"""
    kv_id_fields = ('language', 'digest', 'content_type_id', 'object_id', 'field')
    values = tuple(getattr(self, attr) for attr in kv_id_fields)
    return ('datatrans_%s_%s_%s_%s_%s' % values,
            'datatrans_%s' % self.id)

# cache for an hour
CACHE_DURATION = getattr(settings, 'DATATRANS_CACHE_DURATION', 60 * 60)


class KeyValueManager(models.Manager):
    def get_query_set(self):
        return KeyValueQuerySet(self.model)

    def get_keyvalue(self, key, language, obj, field):
        key = key or ''
        digest = make_digest(key)
        content_type = ContentType.objects.get_for_model(obj.__class__)
        object_id = obj.id
        keyvalue, created = self.get_or_create(digest=digest,
                                               language=language,
                                               content_type_id=content_type.id,
                                               object_id=obj.id,
                                               field=field,
                                               defaults={'value': key})
        return keyvalue

    def lookup(self, key, language, obj, field):
        kv = self.get_keyvalue(key, language, obj, field)
        if kv.edited:
            return kv.value
        else:
            return key

    def for_model(self, model, fields, modelfield=None):
        """
        Get KeyValues for a model. The fields argument is a list of model
        fields.
        If modelfield is specified, only KeyValue entries for that field will
        be returned.
        """
        field_names = [f.name for f in fields] if modelfield is None else [modelfield]
        ct = ContentType.objects.get_for_model(model)
        return self.filter(field__in=field_names, content_type__id=ct.id)

    def contribute_to_class(self, model, name):
        signals.post_save.connect(self._post_save, sender=model)
        signals.post_delete.connect(self._post_delete, sender=model)
        setattr(model, '_get_cache_keys', _get_cache_keys)
        setattr(model, 'cache_keys', property(_get_cache_keys))
        return super(KeyValueManager, self).contribute_to_class(model, name)

    def _invalidate_cache(self, instance):
        """
        Explicitly set a None value instead of just deleting so we don't have
        any race conditions where.
        """
        for key in instance.cache_keys:
            cache.set(key, None, 5)

    def _post_save(self, instance, **kwargs):
        """
        Refresh the cache when saving
        """
        for key in instance.cache_keys:
            cache.set(key, instance, CACHE_DURATION)

    def _post_delete(self, instance, **kwargs):
        self._invalidate_cache(instance)


class KeyValueQuerySet(QuerySet):
    def iterator(self):
        superiter = super(KeyValueQuerySet, self).iterator()
        while True:
            obj = superiter.next()
            # Use cache.add instead of cache.set to prevent race conditions
            for key in obj.cache_keys:
                cache.add(key, obj, CACHE_DURATION)
            yield obj

    def get(self, *args, **kwargs):
        """
        Checks the cache to see if there's a cached entry for this pk. If not,
        fetches using super then stores the result in cache.

        Most of the logic here was gathered from a careful reading of
        ``django.db.models.sql.query.add_filter``
        """
        if self.query.where:
            # If there is any other ``where`` filter on this QuerySet just call
            # super. There will be a where clause if this QuerySet has already
            # been filtered/cloned.
            return super(KeyValueQuerySet, self).get(*args, **kwargs)

        kv_id_fields = ('language', 'digest', 'content_type', 'object_id', 'field')

        # Punt on anything more complicated than get by pk/id only...
        if len(kwargs) == 1:
            k = kwargs.keys()[0]
            if k in ('pk', 'pk__exact', 'id', 'id__exact'):
                obj = cache.get('datatrans_%s' % kwargs.values()[0])
                if obj is not None:
                    return obj
        elif set(kv_id_fields) <= set(kwargs.keys()):
            values = tuple(kwargs[attr] for attr in kv_id_fields)
            obj = cache.get('datatrans_%s_%s_%s_%s_%s' % values)

            if obj is not None:
                return obj

        # Calls self.iterator to fetch objects, storing object in cache.
        return super(KeyValueQuerySet, self).get(*args, **kwargs)


class KeyValue(models.Model):
    """
    The datatrans magic is stored in this model. It stores the localized fields of models.
    """
    content_type = models.ForeignKey(ContentType, null=True)
    object_id = models.PositiveIntegerField(null=True, default=None)
    content_object = generic.GenericForeignKey('content_type', 'object_id')
    field = models.CharField(max_length=255)
    language = models.CharField(max_length=5, db_index=True, choices=settings.LANGUAGES)

    value = models.TextField(blank=True)
    edited = models.BooleanField(blank=True, default=False)
    fuzzy = models.BooleanField(blank=True, default=False)

    digest = models.CharField(max_length=40, db_index=True)
    updated = models.DateTimeField(auto_now=True, default=datetime.datetime.now)

    objects = KeyValueManager()

    def __unicode__(self):
        return u'%s: %s' % (self.language, self.value)

    class Meta:
        #unique_together = ('digest', 'language')
        unique_together = ('language', 'content_type', 'field', 'object_id', 'digest')


class WordCount(models.Model):
    """
    It all happens here
    """
    class Meta:
        abstract = True

    total_words = models.IntegerField(default=0)
    valid = models.BooleanField()


class ModelWordCount(WordCount):
    """
    Caches the total number of localized words for a model
    """
    content_type = models.ForeignKey(ContentType, db_index=True, unique=True)


class FieldWordCount(WordCount):
    """
    Caches the total number of localized words for a model field.
    """
    class Meta:
        unique_together = ('content_type', 'field')

    content_type = models.ForeignKey(ContentType, db_index=True)
    field = models.CharField(max_length=64, db_index=True)





from django.shortcuts import render_to_response
from django.http import HttpResponseRedirect, Http404
from django.core.urlresolvers import reverse
from django.template.context import RequestContext
from django.conf import settings
from django.contrib.auth.decorators import user_passes_test
from django.contrib.contenttypes.models import ContentType

from datatrans import utils
from datatrans.models import KeyValue
from datatrans.utils import count_model_words

def can_translate(user):
    if not user.is_authenticated():
        return False
    elif user.is_superuser:
        return True
    else:
        group_name = getattr(settings, 'DATATRANS_GROUP_NAME', None)
        if group_name:
            from django.contrib.auth.models import Group
            translators = Group.objects.get(name=group_name)
            return translators in user.groups.all()
        else:
            return user.is_staff


def _get_model_slug(model):
    ct = ContentType.objects.get_for_model(model)
    return u'%s.%s' % (ct.app_label, ct.model)


def _get_model_entry(slug):
    app_label, model_slug = slug.split('.')
    ct = ContentType.objects.get(app_label=app_label, model=model_slug)
    model_class = ct.model_class()
    registry = utils.get_registry()
    if not model_class in registry:
        raise Http404(u'No registered model found for given query.')
    return model_class


def _get_model_stats(model, filter=lambda x: x):
    default_lang = utils.get_default_language()
    registry = utils.get_registry()
    keyvalues = filter(KeyValue.objects.for_model(model, registry[model].values()).exclude(language=default_lang))
    total = keyvalues.count()
    done = keyvalues.filter(edited=True, fuzzy=False).count()
    return (done * 100 / total if total > 0 else 0, done, total)


@user_passes_test(can_translate, settings.LOGIN_URL)
def model_list(request):
    """
    Shows an overview of models to translate, along with the fields, languages
    and progress information.
    The context structure is defined as follows:

    context = {'models': [{'languages': [('nl', 'NL', (<percent_done>, <todo>, <total>)), ('fr', 'FR', (<percent_done>, <todo>, <total>))],
                           'field_names': [u'description'],
                           'stats': (75, 15, 20),
                           'slug': u'flags_app.flag',
                           'model_name': u'flag'}]}
    """
    registry = utils.get_registry()

    default_lang = utils.get_default_language()
    languages = [l for l in settings.LANGUAGES if l[0] != default_lang]

    models = [{'slug': _get_model_slug(model),
               'model_name': u'%s' % model._meta.verbose_name,
               'field_names': [u'%s' % f.verbose_name for f in registry[model].values()],
               'stats': _get_model_stats(model),
               'words': count_model_words(model),
               'languages': [
                    (
                        l[0],
                        l[1],
                        _get_model_stats(
                            model,
                            filter=lambda x: x.filter(language=l[0])
                        ),
                    )
                    for l in languages
                ],
               } for model in registry]

    total_words = sum(m['words'] for m in models)
    context = {'models': models, 'words': total_words}

    return render_to_response('datatrans/model_list.html',
                              context,
                              context_instance=RequestContext(request))


def commit_translations(request):
    translations = [
        (KeyValue.objects.get(pk=int(k.split('_')[1])), v)
        for k, v in request.POST.items() if 'translation_' in k]
    for keyvalue, translation in translations:
        empty = 'empty_%d' % keyvalue.pk in request.POST
        ignore = 'ignore_%d' % keyvalue.pk in request.POST
        if translation != '' or empty or ignore:
            if keyvalue.value != translation:
                if not ignore:
                    keyvalue.value = translation
                keyvalue.fuzzy = False
            if ignore:
                keyvalue.fuzzy = False
            keyvalue.edited = True
            keyvalue.save()


def get_context_object(model, fields, language, default_lang, object):
    object_item = {}
    object_item['name'] = unicode(object)
    object_item['id'] = object.id
    object_item['fields'] = object_fields = []
    for field in fields.values():
        key = model.objects.filter(pk=object.pk).values(field.name)[0][field.name]
        original = KeyValue.objects.get_keyvalue(key, default_lang, object, field.name)
        translation = KeyValue.objects.get_keyvalue(key, language, object, field.name)
        object_fields.append({
            'name': field.name,
            'verbose_name': unicode(field.verbose_name),
            'original': original,
            'translation': translation
        })
    return object_item


def needs_translation(model, fields, language, object):
    for field in fields.values():
        key = model.objects.filter(pk=object.pk).values(field.name)[0][field.name]
        translation = KeyValue.objects.get_keyvalue(key, language)
        if not translation.edited:
            return True
    return False


def editor(request, model, language, objects):
    registry = utils.get_registry()
    fields = registry[model]

    default_lang = utils.get_default_language()
    model_name = u'%s' % model._meta.verbose_name

    first_unedited_translation = None
    object_list = []
    for object in objects:
        context_object = get_context_object(
            model, fields, language, default_lang, object)
        object_list.append(context_object)

        if first_unedited_translation is None:
            for field in context_object['fields']:
                tr = field['translation']
                if not tr.edited:
                    first_unedited_translation = tr
                    break

    context = {'model': model_name,
               'objects': object_list,
               'original_language': default_lang,
               'other_language': language,
               'progress': _get_model_stats(
                   model, lambda x: x.filter(language=language)),
               'first_unedited': first_unedited_translation}

    return render_to_response(
        'datatrans/model_detail.html', context,
        context_instance=RequestContext(request))


def selector(request, model, language, objects):
    fields = utils.get_registry()[model]
    for object in objects:
        if needs_translation(model, fields, language, object):
            object.todo = True
    context = {
        'model': model.__name__,
        'objects': objects
    }
    return render_to_response(
        'datatrans/object_list.html', context,
        context_instance=RequestContext(request))


@user_passes_test(can_translate, settings.LOGIN_URL)
def object_detail(request, slug, language, object_id):
    if request.method == 'POST':
        commit_translations(request)
        return HttpResponseRedirect('.')

    model = _get_model_entry(slug)
    objects = model.objects.filter(id=int(object_id))

    return editor(request, model, language, objects)


@user_passes_test(can_translate, settings.LOGIN_URL)
def model_detail(request, slug, language):
    '''
    The context structure is defined as follows:

    context = {'model': '<name of model>',
               'objects': [{'name': '<name of object>',
                            'fields': [{
                                'name': '<name of field>',
                                'original': '<kv>',
                                'translation': '<kv>'
                            ]}],
             }
    '''

    if request.method == 'POST':
        commit_translations(request)
        return HttpResponseRedirect('.')

    model = _get_model_entry(slug)
    meta = utils.get_meta()[model]
    objects = model.objects.all()
    if getattr(meta, 'one_form_per_object', False):
        return selector(request, model, language, objects)
    else:
        return editor(request, model, language, objects)


@user_passes_test(can_translate, settings.LOGIN_URL)
def make_messages(request):
    utils.make_messages()
    return HttpResponseRedirect(reverse('datatrans_model_list'))


@user_passes_test(can_translate, settings.LOGIN_URL)
def obsolete_list(request):
    from django.db.models import Q

    default_lang = utils.get_default_language()
    all_obsoletes = utils.find_obsoletes().order_by('digest')
    obsoletes = all_obsoletes.filter(Q(edited=True) | Q(language=default_lang))

    if request.method == 'POST':
        all_obsoletes.delete()
        return HttpResponseRedirect(reverse('datatrans_obsolete_list'))

    context = {'obsoletes': obsoletes}
    return render_to_response('datatrans/obsolete_list.html', context, context_instance=RequestContext(request))



#!/usr/bin/env python
from distutils.core import setup
import datatrans

LONG_DESCRIPTION = """

"""

setup(name='django-datatrans',
      version=datatrans.__version__,
      description='Translate Django models without changing anything to existing applications and their '
                  'underlying database.',
      long_description=LONG_DESCRIPTION,
      author='Jef Geskens, VikingCo nv',
      author_email='jef.geskens@mobilevikings.com',
      url='http://github.com/citylive/django-datatrans/',
      license='LICENSE.txt',
      packages=['datatrans'],
      include_package_data=True,
      package_data={'datatrans': ['templates/datatrans/*'], },
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: BSD License',
          'Programming Language :: Python',
          'Operating System :: OS Independent',
          'Environment :: Web Environment',
          'Framework :: Django',
          'Topic :: Software Development :: Internationalization',
      ],
)




# -*- encoding: utf-8 -*-
import operator

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured, ObjectDoesNotExist
from django.db import models
from django.utils import translation
from django.utils.datastructures import SortedDict
from django.contrib.contenttypes.models import ContentType

from datatrans.models import KeyValue, make_digest, ModelWordCount, FieldWordCount


"""
REGISTRY is a dict containing the registered models and their translation
fields as a dict.
Example:

>>> from blog.models import Entry
>>> from datatrans.utils import *
>>> class EntryTranslation(object):
...     fields = ('title', 'body',)
...
>>> register(Entry, EntryTranslation)
>>> REGISTRY
{<class 'blog.models.Entry'>: {'body': <django.db.models.fields.TextField object at 0x911368c>,
                               'title': <django.db.models.fields.CharField object at 0x911346c>}}
"""
REGISTRY = SortedDict()
META = SortedDict()


def get_registry():
    return REGISTRY


def get_meta():
    return META


def count_words():
    return sum(count_model_words(model) for model in REGISTRY)


def count_model_words(model):
    """
    Returns word count for the given model and language.
    """
    ct = ContentType.objects.get_for_model(model)
    model_wc, created = ModelWordCount.objects.get_or_create(
        content_type=ct
    )
    if not model_wc.valid:
        total_words = 0

        for field in REGISTRY[model]:
            field_wc, created = FieldWordCount.objects.get_or_create(
                content_type=ct, field=field
            )
            if not field_wc.valid:
                field_wc.total_words = _count_field_words(model, field_wc.field)
                field_wc.valid = True
                field_wc.save()

            total_words += field_wc.total_words

        model_wc.total_words = total_words
        model_wc.valid = True
        model_wc.save()

    return model_wc.total_words


def _count_field_words(model, fieldname):
    """
    Return word count for the given model and field.
    """
    total = 0

    for instance in model.objects.all():
        words = _count_words(instance.__dict__[fieldname])
        total += words
    return total


def _count_words(text):
    """
    Count words in a piece of text.
    """
    return len(text.split()) if text else 0


def get_default_language():
    """
    Get the source language code if specified, or else just the default
    language code.
    """
    lang = getattr(settings, 'SOURCE_LANGUAGE_CODE', settings.LANGUAGE_CODE)
    default = [l[0] for l in settings.LANGUAGES if l[0] == lang]
    if len(default) == 0:
        # when not found, take first part ('en' instead of 'en-us')
        lang = lang.split('-')[0]
        default = [l[0] for l in settings.LANGUAGES if l[0] == lang]
    if len(default) == 0:
        raise ImproperlyConfigured("The [SOURCE_]LANGUAGE_CODE '%s' is not found in your LANGUAGES setting." % lang)
    return default[0]


def get_current_language():
    """
    Get the current language
    """
    lang = translation.get_language()
    current = [l[0] for l in settings.LANGUAGES if l[0] == lang]
    if len(current) == 0:
        lang = lang.split('-')[0]
        current = [l[0] for l in settings.LANGUAGES if l[0] == lang]
    if len(current) == 0:
        # Fallback to default language code
        return get_default_language()
    
    return current[0]


class FieldDescriptor(object):
    def __init__(self, name):
        self.name = name

    def __get__(self, instance, owner):
        lang_code = get_current_language()
        key = instance.__dict__[self.name]
        if not key:
            return u''
        if instance.id is None:
            return key
        return KeyValue.objects.lookup(key, lang_code, instance, self.name)

    def __set__(self, instance, value):
        lang_code = get_current_language()
        default_lang = get_default_language()

        if lang_code == default_lang or not self.name in instance.__dict__ or instance.id is None:
            instance.__dict__[self.name] = value
        else:
            original = instance.__dict__[self.name]
            if original == u'':
                instance.__dict__[self.name] = value
                original = value

            kv = KeyValue.objects.get_keyvalue(original, lang_code, instance, self.name)
            kv.value = value
            kv.edited = True
            kv.save()

        return None


def _pre_save(sender, instance, **kwargs):
    setattr(instance, 'datatrans_old_language', get_current_language())
    default_lang = get_default_language()
    translation.activate(default_lang)

    # When we edit a registered model, update the original translations and mark them as unedited (to do)
    if instance.pk is not None:
        try:
            # Just because instance.pk is set, it does not mean that the instance
            # is saved. Most typical/important ex: loading fixtures
            original = sender.objects.get(pk=instance.pk)
        except ObjectDoesNotExist:
            return None

        ct = ContentType.objects.get_for_model(sender)
        register = get_registry()
        fields = register[sender].values()
        for field in fields:
            old_digest = make_digest(original.__dict__[field.name] or '')
            new_digest = make_digest(instance.__dict__[field.name] or '')
            # If changed, update keyvalues
            if old_digest != new_digest:
                # Check if the new value already exists, if not, create a new one. The old one will be obsoleted.
                old_query = KeyValue.objects.filter(digest=old_digest,
                                                    content_type__id=ct.id,
                                                    object_id=original.id,
                                                    field=field.name)
                new_query = KeyValue.objects.filter(digest=new_digest,
                                                    content_type__id=ct.id,
                                                    object_id=original.id,
                                                    field=field.name)

                old_count = old_query.count()
                new_count = new_query.count()
                _invalidate_word_count(sender, field, instance)
                if old_count != new_count or new_count == 0:
                    for kv in old_query:
                        if new_query.filter(language=kv.language).count() > 0:
                            continue
                        new_value = instance.__dict__[field.name] if kv.language == default_lang else kv.value
                        new_kv = KeyValue(digest=new_digest,
                                          content_type_id=ct.id,
                                          object_id=original.id,
                                          field=field.name,
                                          language=kv.language,
                                          edited=kv.edited,
                                          fuzzy=True,
                                          value=new_value)
                        new_kv.save()


def _post_save(sender, instance, created, **kwargs):
    translation.activate(getattr(instance, 'datatrans_old_language',
                                 get_default_language()))


def _datatrans_filter(self, language=None, mode='and', **kwargs):
    """
    This filter allows you to search model instances on the
    translated contents of a given field.

    :param language: Language code (one of the keys of settings.LANGUAGES).
                     Specifies which language to perform the query in.
    :type language: str

    :param mode: Determines how to combine multiple filter arguments. Mode
                 'or' is fully supported. Mode 'and' only accepts one filter argument.
    :type mode: str, one of ('and', 'or')

    :param <field_name>__<selector>: Indicates which field to search, and
                                     which method (icontains, exact, ...) to use.
                                     If the value of this parameter is an iterable, we will
                                     add multiple filters.
    :type <field_name>__<selector>: str

    :return: Queryset with the matching instances.
    :rtype: :class:`django.db.models.query.QuerySet`

    Search for jobs whose Dutch function name contains 'ontwikkelaar':

    >>> Job.objects.datatrans_filter(function__icontains='ontwikkelaar', language='nl')
    ...

    Search for jobs whose Dutch description contains both 'slim' and 'efficiënt':

    >>> Job.objects.datatrans_filter(description__icontains=['slim', 'efficiënt'],
                                     mode='and', language='nl')
    ...
    """
    assert mode in ('and', 'or')

    if mode == 'and' and len(kwargs) > 1:
        raise NotImplementedError("No support for multiple field name in 'and' mode.")

    if language is None:
        language = translation.get_language()

    registry = get_registry()
    ct = ContentType.objects.get_for_model(self.model)
    q_objects = []

    for key, value in kwargs.items():
        if '__' in key:
            field, method = key.split('__', 1)
        else:
            field, method = key, 'exact'

        if field not in registry[self.model].keys():
            raise ValueError("Field '" + field + "' of " + self.model.__name__ +
                            " has not been registered for translation.")

        def add_filters(field, method, value):
            filters = { 'field' : field, 'value__' + method : value }
            q_objects.append(models.Q(**filters))

        try:
            # Add multiple filters if value is iterable.
            map(lambda v: add_filters(field, method, v), value)
        except TypeError:
            # Iteration failed, therefore value contains single object.
            add_filters(field, method, value)

    query = KeyValue.objects.filter(content_type__id=ct.id, language=language)

    if q_objects:
        op = operator.or_ if mode == 'or' else operator.and_
        query = query.filter(reduce(op, q_objects))

    object_ids = set(i for i , in query.values_list('object_id'))

    return self.filter(id__in=object_ids)


def _invalidate_word_count(model, field, instance):
    content_type = ContentType.objects.get_for_model(model)

    try:
        model_wc = ModelWordCount.objects.get(content_type=content_type)
    except ModelWordCount.DoesNotExist:
        pass
    else:
        model_wc.valid = False
        model_wc.save()

    try:
        field_wc = FieldWordCount.objects.get(
            content_type=content_type, field=field.name
        )
    except FieldWordCount.DoesNotExist:
        pass
    else:
        field_wc.valid = False
        field_wc.save()


def register(model, modeltranslation):
    """
    modeltranslation must be a class with the following attribute:

    fields = ('field1', 'field2', ...)

    For example:

    class BlogPostTranslation(object):
        fields = ('title', 'content',)

    """

    if not model in REGISTRY:
        # create a fields dict (models apparently lack this?!)
        fields = dict([(f.name, f) for f in model._meta._fields() if f.name in modeltranslation.fields])

        REGISTRY[model] = fields
        META[model] = modeltranslation

        models.signals.pre_save.connect(_pre_save, sender=model)
        models.signals.post_save.connect(_post_save, sender=model)

        for field in fields.values():
            setattr(model, field.name, FieldDescriptor(field.name))

        # Set new filter method in model manager.
        model.objects.__class__.datatrans_filter = _datatrans_filter


def make_messages(build_digest_list=False):
    """
    This function loops over all the registered models and, when necessary,
    creates KeyValue entries for the fields specified.

    When build_digest_list is True, a list of digests will be created
    for all the translatable data. When it is False, it will return
    the number of processed objects.
    """
    object_count = 0
    digest_list = []

    for model in REGISTRY:
        fields = REGISTRY[model].values()
        objects = model.objects.all()
        for object in objects:
            for field in fields:
                for lang_code, lang_human in settings.LANGUAGES:
                    value = object.__dict__[field.name]
                    if build_digest_list:
                        digest_list.append(make_digest(value))
                    KeyValue.objects.lookup(value, lang_code, object, field.name)
            object_count += 1

    if build_digest_list:
        return digest_list
    else:
        return object_count


def find_obsoletes():
    digest_list = make_messages(build_digest_list=True)
    obsoletes = KeyValue.objects.exclude(digest__in=digest_list)
    return obsoletes



from django.conf.urls.defaults import *
from datatrans import views

urlpatterns = patterns('',
    url(r'^$', views.model_list, name='datatrans_model_list'),
    url(r'^model/(?P<slug>.*)/(?P<language>.*)/(?P<object_id>.*)/$', views.object_detail, name='datatrans_object_detail'),
    url(r'^model/(?P<slug>.*)/(?P<language>.*)/$', views.model_detail, name='datatrans_model_detail'),
    url(r'^make/messages/$', views.make_messages, name='datatrans_make_messages'),
    url(r'^obsoletes/$', views.obsolete_list, name='datatrans_obsolete_list'),
)
